/*
 * Copyright (c) 1997, 1998  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

#ifndef HAVE_STRCASECMP
#ifdef __STDC__
int strcasecmp(const char *, const char *);
int strncasecmp(const char *, const char *, size_t);
#else /* not __STDC__ */
int strcasecmp()
int strncasecmp();
#endif /* not __STDC__ */
#endif /* not HAVE_STRCASECMP */

#include "eb/eb.h"
#include "eb/error.h"
#include "eb/font.h"

#include "fakelog.h"
#include "getopt.h"
#include "makedir.h"

#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif /* HAVE_STRCHR */

#ifndef HAVE_STRERROR
#ifdef __STDC__
char *strerror(int);
#else /* not __STDC__ */
char *strerror();
#endif /* not __STDC__ */
#endif /* HAVE_STRERROR */

#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX        MAXPATHLEN
#else /* not MAXPATHLEN */
#define PATH_MAX        1024
#endif /* not MAXPATHLEN */
#endif /* not PATH_MAX */

#ifndef O_BINARY
#define O_BINARY 0
#endif


/*
 * Long style options.
 */
static struct option longopts[] = {
  {"debug",             no_argument,       NULL, 'd'},
  {"font-height",       required_argument, NULL, 'f'},
  {"help",              no_argument,       NULL, 'h'},
  {"image-format",      required_argument, NULL, 'i'},
  {"output-directory",  required_argument, NULL, 'o'},
  {"version",           no_argument,       NULL, 'v'},
  {NULL, 0, NULL, 0}
};


/*
 * Supported image formats.
 */
typedef int Image_Format_Code;

#define IMAGE_FORMAT_XBM	0
#define IMAGE_FORMAT_XPM	1
#define IMAGE_FORMAT_GIF	2

typedef struct {
    const char *name;
    const char *suffix;
#ifdef __STDC__
    size_t (*function)(char *, const char *, int, int);
#else
    size_t (*function)();
#endif
} Image_Format;

static Image_Format image_formats[] = {
    {"xbm", "xbm", eb_convert_bitmap_xbm},
    {"xpm", "xpm", eb_convert_bitmap_xpm},
    {"gif", "gif", eb_convert_bitmap_gif},
    {NULL, NULL, NULL}
};

#define MAX_IMAGE_FORMATS	3
#define MAXLEN_IMAGE_NAME	3
#define MAXLEN_IMAGE_SUFFIX	3

/*
 * Generic name of the program.
 */
const char *program_name = "ebfont";

/*
 * Program version.
 */
const char *program_version = VERSION;

/*
 * Actual program name (argv[0])
 */
const char *invoked_name;

/*
 * Debug flag.
 */
static int debug_flag;

/*
 * Defaults and limitations.
 */
#define DEFAULT_FONT_HEIGHT		16
#define DEFAULT_IMAGE_FORMAT		0
#define DEFAULT_OUTPUT_DIRECTORY	"."

#define MAXLEN_FONT_NAME		2
#define MAXLEN_STRING			255


/*
 * Unexported functions.
 */
#ifdef __STDC__
static int parse_image_argument(const char *, int *, Image_Format_Code *);
static int parse_font_argument(const char *, int *, EB_Font_Code *);
static void output_version(void);
static void output_help(void);
static void output_try_help(void);
static int make_fonts(EB_Book *, const char *, EB_Font_Code *, int,
    Image_Format_Code *, int);
static int make_subbook_fonts(EB_Book *, const char *, EB_Font_Code *,
    int, Image_Format_Code *, int);
static int make_subbook_size_fonts(EB_Book *, const char *,
    Image_Format_Code *, int);
static int make_subbook_size_image_fonts(EB_Book *, const char *,
    Image_Format_Code);
static int save_image_file(const char *, const char *, size_t);
static mode_t get_umask(void);
#else
static int parse_image_argument();
static int parse_font_argument();
static void output_version();
static void output_help();
static void output_try_help();
static int make_fonts();
static int make_subbook_fonts();
static int make_subbook_size_fonts();
static int make_subbook_size_image_fonts();
static int save_image_file();
static mode_t get_umask();
#endif


int
main(argc, argv)
    int argc;
    char *argv[];
{
    EB_Book book;
    EB_Font_Code fontlist[EB_MAX_FONTS];
    int fontcount = 0;
    Image_Format_Code imagelist[MAX_IMAGE_FORMATS];
    int imagecount = 0;
    const char *bookpath;
    char outpath[PATH_MAX + 1];
    int make_result;
    int ch;
    
    invoked_name = argv[0];
    debug_flag = 0;
    set_fakelog(FAKELOG_TO_STDERR, LOG_ERR);
    strcpy(outpath, DEFAULT_OUTPUT_DIRECTORY);

    /*
     * Parse command line options.
     */
    for (;;) {
	ch = getopt_long(argc, argv, "df:hi:o:v", longopts, NULL);
	if (ch == EOF)
	    break;
	switch (ch) {
	case 'd':
	    /*
	     * Option `-d'.  Debug mode.
	     */
	    debug_flag = 1;
	    set_fakelog(FAKELOG_TO_STDERR, LOG_DEBUG);
	    break;

	case 'f':
	    /*
	     * Option `-f'.  Generate fonts with HEIGHT.
	     */
	    if (parse_font_argument(optarg, &fontcount, fontlist) < 0)
		exit(1);
	    break;

	case 'h':
	    /*
	     * Option `-h'.  Display help message, then exit.
	     */
	    output_help();
	    exit(0);

	case 'i':
	    /*
	     * Option `-i'.  Generate fonts as FORMAT.
	     */
	    if (parse_image_argument(optarg, &imagecount, imagelist) < 0)
		exit(1);
	    break;

	case 'o':
	    /*
	     * Option `-o'.  Output fonts under DIRECTORY.
	     * The filenames of fonts are:
	     *     "<path>/<subbook>/fonts/<height>/<image>/{narrow,wide}/
	     *         <charno>.<suffix>"
	     *
	     * <path>:    is `optarg'.
	     * <subbook>: require EB_MAXLEN_BASENAME characters.
	     * fonts:     require 5 characters.
	     * <height>:  require 2 characters ("16", "24", "32", or "48").
	     * <image>:   require MAXLEN_IMAGE_NAME characters.
	     * {narrow,wide}:
	     *            requre 6 characters (= strlen("narrow")).
	     * <charno>:  requre 4 characters. ("0000" ... "ffff")
	     * <suffix>:  requre MAXLEN_IMAGE_SUFFIX characters.
	     */
	    if (PATH_MAX < strlen(optarg) + (1 + EB_MAXLEN_BASENAME + 1
		+ 5 + 1 + 2 + 1 + MAXLEN_IMAGE_NAME + 1 + 6 + 1 + 4 + 1
		+ MAXLEN_IMAGE_SUFFIX)) {
		fprintf(stderr, "%s: too long output directory path\n",
		    invoked_name);
		exit(1);
	    }
	    strcpy(outpath, optarg);
	    eb_canonicalize_filename(outpath);
	    break;

	case 'v':
	    /*
	     * Option `-v'.  Display version number, then exit.
	     */
	    output_version();
	    exit(0);

	default:
	    output_try_help();
	    exit(1);
	}
    }

    /*
     * Check the number of rest arguments.
     */
    if (1 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\n", invoked_name);
	output_try_help();
	exit(1);
    }

    if (imagecount == 0) {
	imagelist[0] = DEFAULT_IMAGE_FORMAT;
	imagecount++;
    }

    if (fontcount == 0) {
	fontlist[0] = DEFAULT_FONT_HEIGHT;
	fontcount++;
    }

    /*
     * Generate fonts in a book.
     */
    if (argc == optind)
	bookpath = ".";
    else
	bookpath = argv[optind];

    /*
     * Bind a book.
     */
    if (eb_bind(&book, bookpath) < 0) {
	fprintf(stderr, "%s\n", eb_error_message());
	exit(1);
    }

    /*
     * Make font-files in the book.
     */
    make_result = make_fonts(&book, outpath, fontlist, fontcount, imagelist,
	imagecount);

    /*
     * Dispose the book.
     */
    eb_clear(&book);

    if (make_result < 0)
	exit(1);
    
    return 0;
}


/*
 * Parse an argument to option `--font-height (-f)'
 */
static int
parse_font_argument(argument, fontcount, fontlist)
    const char *argument;
    int *fontcount;
    EB_Font_Code *fontlist;
{
    const char *arg = argument;
    char fontname[MAXLEN_FONT_NAME + 1], *name;
    EB_Font_Code *list;
    EB_Font_Code fontcode;
    int i;

    while (*arg != '\0') {
	/*
	 * Take a next element in the argument.
	 */
	for (i = 0, name = fontname;
	     *arg != ',' && *arg != '\0' && i < MAXLEN_FONT_NAME;
	      i++, name++, arg++)
	    *name = *arg;
	*name = '\0';
	if (*arg == ',')
	    arg++;
	else if (*arg != '\0') {
	    fprintf(stderr, "%s: invalid font height `%s...'\n", invoked_name,
		fontname);
	    return -1;
	}

	/*
	 * Check for the font name.
	 */
	if (strncmp(fontname, "16", 2) == 0)
	    fontcode = 16;
	else if (strncmp(fontname, "24", 2) == 0)
	    fontcode = 24;
	else if (strncmp(fontname, "30", 2) == 0)
	    fontcode = 30;
	else if (strncmp(fontname, "48", 2) == 0)
	    fontcode = 48;
	else {
	    fprintf(stderr, "%s: invalid font height `%s'\n", invoked_name,
		fontname);
	    return -1;
	}

	/*
	 * If the font name is not found in `fontlist', it is added to
	 * the list.
	 */
	for (i = 0, list = fontlist; i < *fontcount; i++, list++) {
	    if (*list == fontcode)
		break;
	}
	if (*fontcount <= i) {
	    *list = fontcode;
	    (*fontcount)++;
	}
    }

    return 0;
}


/*
 * Parse an argument to option `--font-image-format (-i)'
 */
static int
parse_image_argument(argument, imagecount, imagelist)
    const char *argument;
    int *imagecount;
    Image_Format_Code *imagelist;
{
    const char *arg = argument;
    char imagename[MAXLEN_IMAGE_NAME + 1], *name;
    Image_Format *fmt;
    Image_Format_Code *list;
    int imagecode;
    int i;

    while (*arg != '\0') {
	/*
	 * Take a next element in the argument.
	 */
	for (i = 0, name = imagename;
	     *arg != ',' && *arg != '\0' && i < MAXLEN_IMAGE_NAME;
	      i++, name++, arg++)
	    *name = *arg;
	*name = '\0';
	if (*arg == ',')
	    arg++;
	else if (*arg != '\0') {
	    fprintf(stderr, "%s: invalid image format name `%s...'\n",
		invoked_name, imagename);
	    return -1;
	}

	/*
	 * Check for the image-format name.
	 */
	for (i = 0, fmt = image_formats; i < MAX_IMAGE_FORMATS; i++, fmt++) {
	    if (strcasecmp(fmt->name, imagename) == 0)
		break;
	}
	if (MAX_IMAGE_FORMATS <= i) {
	    fprintf(stderr, "%s: invalid image format name `%s'\n",
		invoked_name, imagename);
	    return -1;
	}
	imagecode = i;

	/*
	 * If the image-format name is not found in `imagelist', it
	 * is added to the list.
	 */
	for (i = 0, list = imagelist; i < *imagecount; i++, list++) {
	    if (*list == imagecode)
		break;
	}
	if (*imagecount <= i) {
	    *list = imagecode;
	    (*imagecount)++;
	}
    }

    return 0;
}


/*
 * Output version number to standard error, then exit.
 */
static void
output_version()
{
    fprintf(stderr, "%s version %s\n", program_name, program_version);
}


/*
 * Output help message to standard error, then exit.
 */
static void
output_help()
{
    fprintf(stderr, "Usage: %s [option...] [book-directory]\n", program_name);
    fprintf(stderr, "Options:\n");
    fprintf(stderr, "  -d, --debug                degug mode\n");
    fprintf(stderr, "  -f HEIGHT, --font-height HEIGHT\n");
    fprintf(stderr, "                             generate a font with HEIGHT; 16, 24, 30 or 48\n");
    fprintf(stderr, "                             (default: %d)\n",
	DEFAULT_FONT_HEIGHT);
    fprintf(stderr, "  -h, --help                 display this help, then exit\n");
    fprintf(stderr, "  -i FORMAT, --image-format FORMAT\n");
    fprintf(stderr, "                             generate fonts as FORMAT; xbm, xpm, or gif\n");
    fprintf(stderr, "                             (default: %s)\n",
	image_formats[DEFAULT_IMAGE_FORMAT].name);
    fprintf(stderr, "  -o DIRECTORY, --output-directory DIRECTORY\n");
    fprintf(stderr, "                             output fonts under DIRECTORY\n");
    fprintf(stderr, "                             (default: %s)\n",
	DEFAULT_OUTPUT_DIRECTORY);
    fprintf(stderr, "  -v, --version              display version number, then exit\n");
    fprintf(stderr, "\nArgument:\n");
    fprintf(stderr, "  book-directory             top directory of a CD-ROM book\n");
    fprintf(stderr, "                             (default: .)\n");
    fflush(stderr);
}


/*
 * Output ``try ...'' message to standard error.
 */
static void
output_try_help()
{
    fprintf(stderr, "try `%s --help' for more information\n", invoked_name);
    fflush(stderr);
}


/*
 * Make font-files in the `bookpath'.
 */
static int
make_fonts(book, outpath, fontlist, fontcount, imagelist, imagecount)
    EB_Book *book;
    const char *outpath;
    EB_Font_Code *fontlist;
    int fontcount;
    Image_Format_Code *imagelist;
    int imagecount;
{
    EB_Subbook_Code sublist[EB_MAX_SUBBOOKS];
    int subcount;
    int sub ;
    char subpath[PATH_MAX + 1];
    const char *subdir;

    /*
     * Get a subbook list.
     */
    subcount = eb_subbook_list(book, sublist);
    if (subcount < 0) {
	fprintf(stderr, "%s\n", eb_error_message());
	return -1;
    }


    for (sub = 0; sub < subcount; sub++) {
	/*
	 * Set the current subbook to `sublist[sub]'.
	 */
	if (eb_set_subbook(book, sublist[sub]) < 0) {
	    fprintf(stderr, "%s: subbook=%d\n", eb_error_message(),
		sublist[sub]);
	    return -1;
	}

	/*
	 * Output debug information.
	 */
	if (debug_flag)
	    fprintf(stderr, "debug: subbook %s:\n", subdir);

	/*
	 * Make a directory for the subbook.
	 */
	subdir = eb_subbook_directory(book);
	if (subdir == NULL) {
	    fprintf(stderr, "%s: subbook=%d\n", eb_error_message(),
		sublist[sub]);
	    return -1;
	}
	sprintf(subpath, "%s/%s", outpath, subdir);
	if (make_missing_directory(subpath, 0777 ^ get_umask()) < 0)
	    return -1;

	/*
	 * Make fonts in the subbook.
	 */
	if (make_subbook_fonts(book, subpath, fontlist, fontcount,
	    imagelist, imagecount))
	    return -1;
    }

    return 0;
}


/*
 * Make font-files in the current subbook.
 */
static int
make_subbook_fonts(book, subpath, fontlist, fontcount, imagelist, imagecount)
    EB_Book *book;
    const char *subpath;
    EB_Font_Code *fontlist;
    int fontcount;
    Image_Format_Code *imagelist;
    int imagecount;
{
    const char *subdir;
    char fontpath[PATH_MAX + 1];
    int font;

    /*
     * Get the current subbook name.
     */
    subdir = eb_subbook_directory(book);
    if (subdir == NULL) {
	fprintf(stderr, "%s\n", eb_error_message());
	return -1;
    }

    if (debug_flag)
	fprintf(stderr, "debug: subbook %s:\n", subdir);

    for (font = 0; font < fontcount; font++) {
	/*
	 * Set the current font to `fontlist[font]'.
	 */
	if (!eb_have_font(book, fontlist[font]))
	    continue;
	if (eb_set_font(book, fontlist[font]) < 0) {
	    fprintf(stderr, "%s: subbook=%s, font=%d\n", eb_error_message(),
		subdir, font);
	    return -1;
	}

	/*
	 * Output debug information.
	 */
	if (debug_flag)
	    fprintf(stderr, "debug: font %d:\n", fontlist[font]);

	/*
	 * Make a directory for the font.
	 */
	sprintf(fontpath, "%s/%d", subpath, fontlist[font]);
	if (make_missing_directory(fontpath, 0777 ^ get_umask()) < 0)
	    return -1;

	/*
	 * Make font-files with the size.
	 */
	if (make_subbook_size_fonts(book, fontpath, imagelist, imagecount))
	    return -1;
    }

    return 0;
}


/*
 * Make font-files of the current font.
 */
static int
make_subbook_size_fonts(book, fontpath, imagelist, imagecount)
    EB_Book *book;
    const char *fontpath;
    Image_Format_Code *imagelist;
    int imagecount;
{
    const char *subdir;
    char imagepath[PATH_MAX + 1];
    int image;

    /*
     * Get the current subbook name.
     */
    subdir = eb_subbook_directory(book);
    if (subdir == NULL) {
	fprintf(stderr, "%s\n", eb_error_message());
	return -1;
    }

    for (image = 0; image < imagecount; image++) {
	/*
	 * Output debug information.
	 */
	if (debug_flag) {
	    fprintf(stderr, "debug: image %s:\n",
		image_formats[imagelist[image]].name);
	}

	/*
	 * Make a directory for the image format.
	 */
	sprintf(imagepath, "%s/%s", fontpath,
	    image_formats[imagelist[image]].name);
	if (make_missing_directory(imagepath, 0777 ^ get_umask()) < 0)
	    return -1;

	/*
	 * Make font-files as the image format.
	 */
	if (make_subbook_size_image_fonts(book, imagepath, imagelist[image])
	    < 0)
	    return -1;
    }

    return 0;
}


/*
 * Make font-files of the current font as the image format.
 */
static int
make_subbook_size_image_fonts(book, imagepath, image)
    EB_Book *book;
    const char *imagepath;
    Image_Format_Code image;
{
    const char *subdir;
    char typepath[PATH_MAX + 1];
    char filename[PATH_MAX + 1];
    char bitmap[EB_SIZE_WIDE_FONT_48];
    char imagedata[EB_SIZE_FONT_IMAGE];
    size_t imagesize;
    int width, height;
    int ch;

    /*
     * Get the current subbook name.
     */
    subdir = eb_subbook_directory(book);
    if (subdir == NULL) {
	fprintf(stderr, "%s\n", eb_error_message());
	return -1;
    }

    /*
     * Get the current font size.
     */
    height = eb_font(book);
    if (height < 0) {
	fprintf(stderr, "%s: subbook=%s\n", eb_error_message(), subdir);
	return -1;
    }

    /*
     * Make the bitmap files of narrow fonts.
     */
    if (eb_have_narrow_font(book)) {
	/*
	 * Get narrow font information.
	 */
	width = eb_narrow_font_width(book);
	if (width < 0) {
	    fprintf(stderr, "%s: subbook=%s, font=%d, type=narrow\n",
		eb_error_message(), subdir, height);
	    return -1;
	}
	ch = eb_narrow_font_start(book);
	if (ch < 0) {
	    fprintf(stderr, "%s: subbook=%s, font=%d, type=narrow\n",
		eb_error_message(), subdir, height);
	    return -1;
	}

	/*
	 * Make a directory for the narrow font.
	 */
	sprintf(typepath, "%s/narrow", imagepath);
	if (make_missing_directory(typepath, 0777 ^ get_umask()) < 0)
	    return -1;

	while (0 <= ch) {
	    /*
	     * Output debug information.
	     */
	    if (debug_flag)
		fprintf(stderr, "debug: character %04x\n", ch);

	    /*
	     * Generate a bitmap file for the character `ch'.
	     */
	    sprintf(filename, "%s/%04x.%s", typepath, ch,
		image_formats[image].suffix);
	    if (eb_narrow_character_bitmap(book, ch, bitmap) < 0) {
		fprintf(stderr, "%s: subbook=%s, font=%d, type=narrow, \
character=0x%04x\n",
		    eb_error_message(), subdir, height, ch);
		return -1;
	    }
	    imagesize = (image_formats[image].function)(imagedata, bitmap,
		width, height);
	    if (save_image_file(filename, imagedata, imagesize) < 0)
		return -1;

	    /*
	     * Toward next charahacter.
	     */
	    ch = eb_forward_narrow_character(book, ch, 1);
	}
    }

    /*
     * Make the bitmap files of wide fonts.
     */
    if (eb_have_wide_font(book)) {
	/*
	 * Get wide font information.
	 */
	width = eb_wide_font_width(book);
	if (width < 0) {
	    fprintf(stderr, "%s: subbook=%s, font=%d, type=wide\n",
		eb_error_message(), subdir, height);
	    return -1;
	}
	ch = eb_wide_font_start(book);
	if (ch < 0) {
	    fprintf(stderr, "%s: subbook=%s, font=%d, type=wide\n",
		eb_error_message(), subdir, height);
	    return -1;
	}

	/*
	 * Make a directory for the wide font.
	 */
	sprintf(typepath, "%s/wide", imagepath);
	if (make_missing_directory(typepath, 0777 ^ get_umask()) < 0)
	    return -1;

	while (0 <= ch) {
	    /*
	     * Output debug information.
	     */
	    if (debug_flag)
		fprintf(stderr, "debug: character %04x\n", ch);

	    /*
	     * Generate a bitmap file for the character `ch'.
	     */
	    sprintf(filename, "%s/%04x.%s", typepath, ch,
		image_formats[image].suffix);
	    if (eb_wide_character_bitmap(book, ch, bitmap) < 0) {
		fprintf(stderr, "%s: subbook=%s, font=%d, type=wide, \
character=0x%04x\n",
		    eb_error_message(), subdir, height, ch);
		return -1;
	    }
	    imagesize = (image_formats[image].function)(imagedata, bitmap,
		width, height);
	    if (save_image_file(filename, imagedata, imagesize) < 0)
		return -1;

	    /*
	     * Toward next charahacter.
	     */
	    ch = eb_forward_wide_character(book, ch, 1);
	}
    }

    return 0;
}


/*
 * Save an image file.
 */
static int
save_image_file(filename, imagedata, imagesize)
    const char *filename;
    const char *imagedata;
    size_t imagesize;
{
    int file = -1;

#ifdef O_CREAT
    file = open(filename, O_CREAT | O_TRUNC | O_WRONLY | O_BINARY,
	0666 ^ get_umask());
#else
    file = creat(filename, 0666 ^ get_umask());
#endif

    if (file < 0) {
	fprintf(stderr, "cannot open the file, %s: %s\n", strerror(errno),
	    filename);
	goto failed;
    }
    if (write(file, imagedata, imagesize) != imagesize) {
	fprintf(stderr, "cannot write to the file, %s: %s\n", strerror(errno),
	    filename);
	goto failed;
    }
    if (close(file) < 0) {
	fprintf(stderr, "cannot write to the file, %s: %s\n", strerror(errno),
	    filename);
	file = -1;
	goto failed;
    }

    return 0;

    /*
     * An error occurs...
     */
  failed:
    if (0 <= file && close(file) < 0) {
	fprintf(stderr, "cannot close the file, %s: %s\n", strerror(errno),
	    filename);
    }

    return -1;
}


/*
 * Get the current umask.
 */
static mode_t
get_umask()
{
    mode_t um;

    um = umask(022);
    umask(um);

    return um;
}


